"""Unit tests for djblets.extensions.hooks.DataGridColumnsHook."""

from __future__ import unicode_literals

from kgb import SpyAgency

from djblets.datagrid.grids import Column, DataGrid
from djblets.extensions.extension import Extension
from djblets.extensions.hooks import DataGridColumnsHook
from djblets.extensions.tests.base import ExtensionTestsMixin
from djblets.testing.testcases import TestCase


class TestExtension(Extension):
    pass


class DataGridColumnsHookTest(SpyAgency, ExtensionTestsMixin, TestCase):
    """Unit tests for djblets.extensions.hooks.DataGridColumnsHook."""

    def setUp(self):
        super(DataGridColumnsHookTest, self).setUp()

        self.extension = self.setup_extension(TestExtension)

    def test_add_column(self):
        """Testing DataGridColumnsHook registers column"""
        self.spy_on(DataGrid.add_column)

        DataGridColumnsHook(extension=self.extension,
                            datagrid_cls=DataGrid,
                            columns=[Column(id='sandbox')])

        self.assertTrue(DataGrid.add_column.called)

    def test_remove_column(self):
        """Testing DataGridColumnsHook unregisters column"""
        self.spy_on(DataGrid.remove_column)

        hook = DataGridColumnsHook(extension=self.extension,
                                   datagrid_cls=DataGrid,
                                   columns=[Column(id='sandbox2')])

        hook.disable_hook()

        self.assertTrue(DataGrid.remove_column.called)



"""Unit tests for djblets.auth.util."""

from __future__ import unicode_literals

from django import forms
from django.contrib.auth.models import User
from django.contrib.sessions.middleware import SessionMiddleware
from django.test.client import RequestFactory

from djblets.auth.util import validate_old_password, validate_test_cookie
from djblets.testing.testcases import TestCase


class ValidateOldPasswordTests(TestCase):
    """Unit tests for djblets.auth.util.validate_old_password."""

    class DummyForm(forms.Form):
        password = forms.CharField(required=False)
        alt_password = forms.CharField(required=False)

    def setUp(self):
        super(ValidateOldPasswordTests, self).setUp()

        self.user = User.objects.create_user(username='test-user',
                                             email='user@example.com',
                                             password='password')

    def test_with_password_matches(self):
        """Testing validate_old_password with password matching User's
        current password
        """
        form = self.DummyForm({
            'password': 'password',
        })
        validate_old_password(form, self.user)

        self.assertNotIn('password', form.errors)

    def test_with_password_does_not_match(self):
        """Testing validate_old_password with password not matching User's
        current password
        """
        form = self.DummyForm({
            'password': 'bad-password',
        })
        validate_old_password(form, self.user)

        self.assertIn('password', form.errors)
        self.assertEqual(form.errors['password'], ['Incorrect password.'])

    def test_with_existing_errors(self):
        """Testing validate_old_password with field already containing errors
        """
        form = self.DummyForm({
            'password': 'bad-password',
        })
        form.errors['password'] = form.error_class(['Something went wrong.'])
        validate_old_password(form, self.user)

        self.assertIn('password', form.errors)
        self.assertEqual(form.errors['password'], ['Something went wrong.'])

    def test_with_field_name_and_password_matches(self):
        """Testing validate_old_password with explicit field_name and
        password matching User's current password
        """
        form = self.DummyForm({
            'alt_password': 'password',
        })
        validate_old_password(form, self.user, field_name='alt_password')

        self.assertNotIn('alt_password', form.errors)

    def test_with_field_name_and_password_does_not_match(self):
        """Testing validate_old_password with explicit field name and password
        not matching User's current password
        """
        form = self.DummyForm({
            'alt_password': 'bad-password',
        })
        validate_old_password(form, self.user, field_name='alt_password')

        self.assertIn('alt_password', form.errors)
        self.assertEqual(form.errors['alt_password'], ['Incorrect password.'])

    def test_with_field_name_and_existing_errors(self):
        """Testing validate_old_password with explicit field name and field
        already containing errors
        """
        form = self.DummyForm({
            'alt_password': 'bad-password',
        })
        form.errors['alt_password'] = form.error_class(
            ['Something went wrong.'])
        validate_old_password(form, self.user, field_name='alt_password')

        self.assertIn('alt_password', form.errors)
        self.assertEqual(form.errors['alt_password'],
                         ['Something went wrong.'])


class ValidateTestCookieTests(TestCase):
    """Unit tests for djblets.auth.util.validate_test_cookie."""

    def setUp(self):
        super(ValidateTestCookieTests, self).setUp()

        class DummyForm(forms.Form):
            pass

        self.form = DummyForm()

        self.request = RequestFactory().post('/')
        SessionMiddleware().process_request(self.request)

    def test_with_test_cookie(self):
        """Testing validate_test_cookie with test cookie set"""
        self.request.session.set_test_cookie()

        validate_test_cookie(self.form, self.request)

        self.assertNotIn('submit', self.form.errors)

    def test_without_test_cookie(self):
        """Testing validate_test_cookie without test cookie set"""
        validate_test_cookie(self.form, self.request)

        self.assertIn('submit', self.form.errors)
        self.assertEqual(self.form.errors['submit'],
                         ['Cookies must be enabled.'])



from __future__ import unicode_literals

from django import forms

from djblets.forms.forms import KeyValueForm
from djblets.testing.testcases import TestCase


class DummyForm(KeyValueForm):
    char_field = forms.CharField(initial='default', required=False)
    bool_field = forms.BooleanField(initial=True, required=False)

    def __init__(self, *args, **kwargs):
        self.saved = False

        super(DummyForm, self).__init__(*args, **kwargs)

    def create_instance(self):
        return {}

    def save_instance(self):
        self.saved = True


class KeyValueFormTests(TestCase):
    """Unit tests for djblets.forms.forms.KeyValueForm."""

    def test_load_without_instance(self):
        """Testing KeyValueForm load without instance"""
        form = DummyForm()

        self.assertEqual(form.fields['char_field'].initial, 'default')
        self.assertTrue(form.fields['bool_field'].initial)

    def test_load_with_instance(self):
        """Testing KeyValueForm.load with instance"""
        form = DummyForm(instance={
            'char_field': 'new value',
            'bool_field': False,
        })

        self.assertEqual(form.fields['char_field'].initial, 'new value')
        self.assertFalse(form.fields['bool_field'].initial)

    def test_load_with_load_blacklist(self):
        """Testing KeyValueForm.load with Meta.load_blacklist"""
        class LoadBlacklistForm(DummyForm):
            class Meta:
                load_blacklist = ('char_field',)
                save_blacklist = ('bool_field',)

        form = LoadBlacklistForm(instance={
            'char_field': 'new value',
            'bool_field': False,
        })

        self.assertEqual(form.fields['char_field'].initial, 'default')
        self.assertFalse(form.fields['bool_field'].initial)

    def test_load_with_save_blacklist(self):
        """Testing KeyValueForm.load with Meta.save_blacklist only"""
        class LoadBlacklistForm(DummyForm):
            class Meta:
                save_blacklist = ('char_field',)

        form = LoadBlacklistForm(instance={
            'char_field': 'new value',
            'bool_field': False,
        })

        self.assertEqual(form.fields['char_field'].initial, 'default')
        self.assertFalse(form.fields['bool_field'].initial)

    def test_load_with_disabled_fields(self):
        """Testing KeyValueForm.load with disabled_fields"""
        class DisabledFieldsForm(DummyForm):
            def load(self):
                self.disabled_fields['char_field'] = True

                super(DisabledFieldsForm, self).load()

            class Meta:
                save_blacklist = ('char_field',)

        form = DisabledFieldsForm(instance={
            'char_field': 'new value',
            'bool_field': False,
        })

        self.assertIn('disabled', form.fields['char_field'].widget.attrs)
        self.assertEqual(form.fields['char_field'].widget.attrs['disabled'],
                         'disabled')
        self.assertNotIn('disabled', form.fields['bool_field'].widget.attrs)

    def test_load_with_custom_deserialized_field(self):
        """Testing KeyValueForm.load with custom deserializer for field"""
        class DeserializerForm(DummyForm):
            custom = forms.CharField()

            def deserialize_custom_field(self, value):
                return value['value']

        form = DeserializerForm(instance={
            'custom': {
                'is_custom': True,
                'value': 'my value',
            },
        })

        self.assertEqual(form.fields['custom'].initial, 'my value')

    def test_save_without_instance(self):
        """Testing KeyValueForm.save without existing instance"""
        form = DummyForm(data={
            'char_field': 'new value',
            'bool_field': False,
        })

        self.assertTrue(form.is_valid())
        result = form.save()

        self.assertEqual(
            result,
            {
                'char_field': 'new value',
                'bool_field': False,
            })
        self.assertTrue(form.saved)

    def test_save_with_instance(self):
        """Testing KeyValueForm.save with existing instance"""
        instance = {
            'char_field': 'orig value',
            'bool_field': False,
        }
        form = DummyForm(
            instance=instance,
            data={
                'char_field': 'new value',
                'bool_field': False,
            })

        self.assertTrue(form.is_valid())
        result = form.save()

        self.assertEqual(
            result,
            {
                'char_field': 'new value',
                'bool_field': False,
            })
        self.assertTrue(form.saved)
        self.assertTrue(instance is result)

    def test_save_with_instance_no_commit(self):
        """Testing KeyValueForm.save with existing instance and with
        commit=False
        """
        instance = {
            'char_field': 'orig value',
            'bool_field': False,
        }
        form = DummyForm(
            instance=instance,
            data={
                'char_field': 'new value',
                'bool_field': False,
            })

        self.assertTrue(form.is_valid())
        result = form.save(commit=False)

        self.assertEqual(
            result,
            {
                'char_field': 'new value',
                'bool_field': False,
            })
        self.assertFalse(form.saved)
        self.assertTrue(instance is result)

    def test_save_with_save_blacklist(self):
        """Testing KeyValueForm.save with Meta.save_blacklist"""
        class SaveBlacklistForm(DummyForm):
            class Meta:
                save_blacklist = ('char_field',)

        instance = {
            'char_field': 'orig value',
            'bool_field': False,
        }
        form = SaveBlacklistForm(
            instance=instance,
            data={
                'char_field': 'new value',
                'bool_field': True,
            })

        self.assertTrue(form.is_valid())
        result = form.save()

        self.assertEqual(
            result,
            {
                'char_field': 'orig value',
                'bool_field': True,
            })
        self.assertTrue(instance is result)

    def test_save_with_extra_save_blacklist(self):
        """Testing KeyValueForm.save with Meta.extra_save_blacklist"""
        class SaveBlacklistForm(DummyForm):
            class Meta:
                save_blacklist = ('char_field',)

        instance = {
            'char_field': 'orig value',
            'bool_field': False,
        }
        form = SaveBlacklistForm(
            instance=instance,
            data={
                'char_field': 'new value',
                'bool_field': True,
            })

        self.assertTrue(form.is_valid())
        result = form.save(extra_save_blacklist=('bool_field',))

        self.assertEqual(
            result,
            {
                'char_field': 'orig value',
                'bool_field': False,
            })
        self.assertTrue(instance is result)

    def test_save_with_custom_serialized_field(self):
        """Testing KeyValueForm.save with custom serializer for field"""
        class SerializerForm(DummyForm):
            custom = forms.CharField()

            def serialize_custom_field(self, value):
                return {
                    'is_custom': True,
                    'value': value,
                }

        instance = {
            'custom': {
                'is_custom': True,
                'value': 'orig value',
            },
        }

        form = SerializerForm(
            instance=instance,
            data={
                'custom': 'new value',
            })

        self.assertTrue(form.is_valid())
        result = form.save()

        self.assertTrue(instance is result)
        self.assertEqual(
            result['custom'],
            {
                'is_custom': True,
                'value': 'new value',
            })



"""Unit tests for djblets.extensions.hooks.SignalHook."""

from __future__ import unicode_literals

from django.dispatch import Signal
from kgb import SpyAgency

from djblets.extensions.extension import Extension
from djblets.extensions.hooks import SignalHook
from djblets.extensions.tests.base import ExtensionTestsMixin
from djblets.testing.testcases import TestCase


class TestExtension(Extension):
    pass


class SignalHookTests(SpyAgency, ExtensionTestsMixin, TestCase):
    """Unit tests for djblets.extensions.hooks.SignalHook."""

    def setUp(self):
        super(SignalHookTests, self).setUp()

        self.test_extension = self.setup_extension(TestExtension)

        self.signal = Signal()
        self.spy_on(self._on_signal_fired)
        self.spy_on(self._on_signal_exception)

    def test_initialize(self):
        """Testing SignalHook initialization connects to signal"""
        SignalHook(self.test_extension, self.signal, self._on_signal_fired)

        self.assertEqual(len(self._on_signal_fired.calls), 0)
        self.signal.send(self)
        self.assertEqual(len(self._on_signal_fired.calls), 1)

    def test_shutdown(self):
        """Testing SignalHook.shutdown disconnects from signal"""
        hook = SignalHook(self.test_extension, self.signal,
                          self._on_signal_fired)
        hook.disable_hook()

        self.assertEqual(len(self._on_signal_fired.calls), 0)
        self.signal.send(self)
        self.assertEqual(len(self._on_signal_fired.calls), 0)

    def test_shutdown_with_sender(self):
        """Testing SignalHook.shutdown disconnects when a sender was set"""
        hook = SignalHook(self.test_extension, self.signal,
                          self._on_signal_fired, sender=self)
        hook.disable_hook()

        self.assertEqual(len(self._on_signal_fired.calls), 0)
        self.signal.send(self)
        self.assertEqual(len(self._on_signal_fired.calls), 0)

    def test_forwards_args(self):
        """Testing SignalHook forwards arguments to callback"""
        seen_kwargs = {}

        def callback(**kwargs):
            seen_kwargs.update(kwargs)

        SignalHook(self.test_extension, self.signal, callback)
        self.signal.send(sender=self, foo=1, bar=2)

        self.assertTrue('foo', seen_kwargs)
        self.assertEqual(seen_kwargs['foo'], 1)
        self.assertTrue('bar', seen_kwargs)
        self.assertEqual(seen_kwargs['bar'], 2)

    def test_sandbox_errors_true(self):
        """Testing SignalHook with sandbox_errors set to True logs errors"""
        SignalHook(self.test_extension, self.signal, self._on_signal_exception,
                   sandbox_errors=True)

        self.assertEqual(len(self._on_signal_exception.calls), 0)
        self.signal.send(self)
        self.assertEqual(len(self._on_signal_exception.calls), 1)

    def test_sandbox_errors_false(self):
        """Testing SignalHook with sandbox_errors set to False"""
        SignalHook(self.test_extension, self.signal, self._on_signal_exception,
                   sandbox_errors=False)

        self.assertEqual(len(self._on_signal_exception.calls), 0)
        self.assertRaises(Exception, self.signal.send, self)
        self.assertEqual(len(self._on_signal_exception.calls), 1)

    def _on_signal_fired(self, *args, **kwargs):
        pass

    def _on_signal_exception(self, *args, **kwargs):
        raise Exception



"""Mixin for class-based views that support ETags."""

from __future__ import unicode_literals

from django.http import HttpResponseNotModified

from djblets.util.http import encode_etag, set_etag, etag_if_none_match


class ETagViewMixin(object):
    """Mixin to handle ETag generation for a page.

    This enables ETag support in the view for HTTP GET requests. On a request,
    :py:meth:`get_etag_data` will be called, which will calculate an ETag for
    the content. If the client already knows this ETag, it will receive a
    :http:`304`. Otherwise, it will return a response with that ETag attached.

    This currently only supports ETags for HTTP GET and HEAD methods.
    """

    def dispatch(self, request, *args, **kwargs):
        """Handle a HTTP request and dispatch to a handler.

        Args:
            request (django.http.HttpRequest):
                The HTTP request from the client.

            *args (tuple):
                Positional arguments passed to the handler.

            **kwargs (dict):
                Keyword arguments passed to the handler.

        Returns:
            django.http.HttpResponse:
            The resulting HTTP response from the handler.
        """
        handle_etag = request.method.upper() in ('GET', 'HEAD')

        if handle_etag:
            etag = self.get_etag_data(request, *args, **kwargs)

            if etag:
                etag = encode_etag(etag)

                if etag_if_none_match(request, etag):
                    return HttpResponseNotModified()

        response = super(ETagViewMixin, self).dispatch(request, *args,
                                                       **kwargs)

        if handle_etag and etag and 200 <= response.status_code < 300:
            set_etag(response, etag)

        return response



from __future__ import unicode_literals

import inspect
from functools import wraps

from django.contrib.auth.models import User
from django.utils import six
from django.utils.decorators import method_decorator


def add_fixtures(fixtures, replace=False):
    """Adds or replaces the fixtures used for this test.

    This must be used along with :py:func:`djblets.testing.testcases.TestCase`.
    """
    def _dec(func):
        func._fixtures = fixtures
        func._replace_fixtures = replace
        return func

    return _dec


def requires_user_profile(f):
    """Ensure that User.get_profile exists for the following unit test.

    In Django 1.7+, ``User.get_profile`` no longer exists. Currently some
    parts of Djblets require this method to exist in Django or be added by the
    consuming app. To fix this requirement for unit testing on Django 1.7+, we
    attach a method that raises :py:exc:`NotImplementedError` when the method
    does not exist. After the test exits the method will be removed.

    Args:
        f (callable or type):
            The function or class to decorate

    Returns:
        callable or type:
        The decorated function or class.
    """
    def get_profile(self):
        raise NotImplementedError

    if inspect.isclass(f):
        # f is a class, so we will decorate all test_ methods on the class.
        decorator = method_decorator(requires_user_profile)
        attrs = vars(f)

        for attr_name, value in six.iteritems(attrs):
            if attr_name.startswith('test_') and callable(value):
                setattr(f, attr_name, decorator(value))

        return f
    else:
        @wraps(f)
        def decorated(*args, **kwargs):
            add_profile = not hasattr(User, 'get_profile')

            if add_profile:
                setattr(User, 'get_profile', get_profile)

            try:
                f(*args, **kwargs)
            finally:
                # Ensure that the method is the one we added. It is possible
                # that some code has replaced the method.
                if add_profile and hasattr(User, 'get_profile'):
                    # We have to un-spy for KGB-added spies because they will
                    # end up re-adding User.get_profile in test tear down after
                    # this decorator exits otherwise.
                    if hasattr(User.get_profile, 'unspy'):
                        User.get_profile.unspy()

                    # On Python 3.x, unbound methods are just plain functions,
                    # so they don't have __func__.
                    if hasattr(User.get_profile, '__func__'):
                        # Python 2.x
                        user_get_profile = User.get_profile.__func__
                    else:
                        # Python 3.x
                        user_get_profile = User.get_profile

                    if user_get_profile is get_profile:
                        delattr(User, 'get_profile')
        return decorated



from __future__ import unicode_literals

import json
from collections import OrderedDict

from djblets.testing.testcases import TestCase
from djblets.webapi.encoders import (JSONEncoderAdapter, WebAPIEncoder,
                                     XMLEncoderAdapter)


class EncoderAdapterTests(TestCase):
    """Tests encoding correctness of WebAPIEncoder adapters"""

    @classmethod
    def setUpClass(cls):
        super(EncoderAdapterTests, cls).setUpClass()

        data = OrderedDict()
        data['string_val'] = 'foobar'
        data['none_val'] = None
        data['dict_val'] = {'foo': 'bar'}
        data['bool_val'] = True
        data['scientific_val'] = 2.75e-15
        data['int_val'] = 42
        data['float_val'] = 3.14159
        data['list_val'] = [10, 'baz']

        cls.data = data

    def test_json_encoder_adapter(self):
        """Testing JSONEncoderAdapter.encode"""
        encoder = WebAPIEncoder()
        adapter = JSONEncoderAdapter(encoder)

        content = adapter.encode(self.data)
        self.assertEqual(content, json.dumps(self.data, sort_keys=True))

    def test_xml_encoder_adapter(self):
        """Testing XMLEncoderAdapter.encode"""
        encoder = WebAPIEncoder()
        adapter = XMLEncoderAdapter(encoder)

        expected = (
            '<?xml version="1.0" encoding="utf-8"?>\n'
            '<rsp>\n'
            ' <string_val>foobar</string_val>\n'
            ' <none_val>\n'
            ' </none_val>\n'
            ' <dict_val>\n'
            '  <foo>bar</foo>\n'
            ' </dict_val>\n'
            ' <bool_val>1</bool_val>\n'
            ' <scientific_val>2.75e-15</scientific_val>\n'
            ' <int_val>42</int_val>\n'
            ' <float_val>3.14159</float_val>\n'
            ' <list_val>\n'
            '  <array>\n'
            '   <item>10</item>\n'
            '   <item>baz</item>\n'
            '  </array>\n'
            ' </list_val>\n'
            '</rsp>'
        )

        content = adapter.encode(self.data)
        self.assertEqual(content, expected)



"""Unit tests for djblets.util.templatetags.djblets_images."""

from __future__ import unicode_literals

from collections import OrderedDict

from django.template import Context, Template, TemplateSyntaxError

from djblets.testing.testcases import TestCase


class SrcSetTagTests(TestCase):
    """Unit tests for the {% srcset %} template tag."""

    def test_with_valid_sources(self):
        """Testing {% srcset %} with valid sources"""
        t = Template('{% load djblets_images %}'
                     '<img srcset="{% srcset sources %}">')

        self.assertEqual(
            t.render(Context({
                'sources': {
                    '1x': '/path/to/image.png',
                    '2x': '/path/to/image@2x.png',
                    '3x': '/path/to/image@3x.png',
                    '10w': '/path/to/image@10w.png',
                },
            })),
            '<img srcset="'
            '/path/to/image.png 1x, '
            '/path/to/image@2x.png 2x, '
            '/path/to/image@3x.png 3x, '
            '/path/to/image@10w.png 10w">')

    def test_with_invalid_sources(self):
        """Testing {% srcset %} with invalid sources"""
        t = Template('{% load djblets_images %}'
                     '<img srcset="{% srcset sources %}">')

        message = '"ax" is not a valid srcset size descriptor.'

        with self.assertRaisesMessage(TemplateSyntaxError, message):
            t.render(Context({
                'sources': {
                    'ax': '/path/to/image.png',
                },
            }))

    def test_sorts(self):
        """Testing {% srcset %} sorts sources"""
        t = Template('{% load djblets_images %}'
                     '<img srcset="{% srcset sources %}">')

        sources = OrderedDict()
        sources['100w'] = '/path/to/image@100w.png'
        sources['10w'] = '/path/to/image@10w.png'
        sources['3x'] = '/path/to/image@3x.png'
        sources['1x'] = '/path/to/image.png'
        sources['2x'] = '/path/to/image@2x.png'
        sources['20w'] = '/path/to/image@20w.png'
        sources['1.5x'] = '/path/to/image@1.5x.png'
        sources['2w'] = '/path/to/image@2w.png'
        sources['10x'] = '/path/to/image@10x.png'

        self.assertEqual(
            t.render(Context({
                'sources': sources,
            })),
            '<img srcset="'
            '/path/to/image.png 1x, '
            '/path/to/image@1.5x.png 1.5x, '
            '/path/to/image@2x.png 2x, '
            '/path/to/image@3x.png 3x, '
            '/path/to/image@10x.png 10x, '
            '/path/to/image@2w.png 2w, '
            '/path/to/image@10w.png 10w, '
            '/path/to/image@20w.png 20w, '
            '/path/to/image@100w.png 100w">')


class ImageSourceAttrsTagTests(TestCase):
    """Unit tests for the {% image_source_attrs %} template tag."""

    def test_with_valid_sources(self):
        """Testing {% image_source_attrs %} with valid sources"""
        t = Template(
            '{% load djblets_images %}'
            '<img {% image_source_attrs %}\n'
            '      1x /path/to/image.png\n'
            '      2x /path/to/image@2x.png\n'
            '      3x /path/to/image@3x.png\n'
            '      10w /path/to/image@10w.png\n'
            '     {% end_image_source_attrs %}>')

        self.assertEqual(
            t.render(Context()),
            '<img src="/path/to/image.png" srcset="'
            '/path/to/image.png 1x, '
            '/path/to/image@2x.png 2x, '
            '/path/to/image@3x.png 3x, '
            '/path/to/image@10w.png 10w">')

    def test_with_invalid_sources(self):
        """Testing {% image_source_attrs %} with invalid sources definition"""
        t = Template(
            '{% load djblets_images %}'
            '<img {% image_source_attrs %}\n'
            '      /path/to/image.png\n'
            '      2x /path/to/image@2x.png\n'
            '     {% end_image_source_attrs %}>')

        message = (
            'The source definition passed to {% image_source_attrs %} is '
            'not structured correctly. Make sure that there is one source '
            'definition per line and that it contains a descriptor and a '
            'URL.'
        )

        with self.assertRaisesMessage(TemplateSyntaxError, message):
            t.render(Context())

    def test_with_no_1x(self):
        """Testing {% image_source_attrs %} with no 1x source"""
        t = Template(
            '{% load djblets_images %}'
            '<img {% image_source_attrs %}\n'
            '      2x /path/to/image@2x.png\n'
            '      3x /path/to/image@3x.png\n'
            '     {% end_image_source_attrs %}>')

        message = (
            'The source definition passed to {% image_source_attr %} must '
            'contain a "1x" descriptor.'
        )

        with self.assertRaisesMessage(TemplateSyntaxError, message):
            t.render(Context())



"""Unit tests for djblets.util.templatetags.djblets_utils."""

from __future__ import unicode_literals

import warnings
from collections import OrderedDict
from datetime import datetime, timedelta

from django.http import HttpRequest, QueryDict
from django.template import Context, Template
from django.utils import six

from djblets.deprecation import RemovedInDjblets30Warning
from djblets.testing.testcases import TagTest, TestCase
from djblets.util.templatetags.djblets_utils import (ageid, escapespaces,
                                                     humanize_list, indent)


class AgeIdTagTests(TagTest):
    """Unit tests for the {% ageid %} template tag."""

    def setUp(self):
        super(AgeIdTagTests, self).setUp()

        self.now = datetime.utcnow()

        self.context = {
            'now': self.now,
            'minus1': self.now - timedelta(1),
            'minus2': self.now - timedelta(2),
            'minus3': self.now - timedelta(3),
            'minus4': self.now - timedelta(4),
        }

    def test_with_now(self):
        """Testing {% ageid %} with now"""
        self.assertEqual(ageid(self.now), 'age1')

    def test_with_now_minus_1_day(self):
        """Testing {% ageid %} with yesterday"""
        self.assertEqual(ageid(self.now - timedelta(1)), 'age2')

    def test_with_now_minus_2_days(self):
        """Testing {% ageid %} with two days ago"""
        self.assertEqual(ageid(self.now - timedelta(2)), 'age3')

    def test_with_now_minus_3_days(self):
        """Testing {% ageid %} with three days ago"""
        self.assertEqual(ageid(self.now - timedelta(3)), 'age4')

    def test_with_now_minus_4_days(self):
        """Testing {% ageid %} with four days ago"""
        self.assertEqual(ageid(self.now - timedelta(4)), 'age5')

    def test_with_non_datetime(self):
        """Testing {% ageid %} with non-datetime object"""
        class Foo:
            def __init__(self, now):
                self.day = now.day
                self.month = now.month
                self.year = now.year

        self.assertEqual(ageid(Foo(self.now)), 'age1')


class AttrTagTests(TestCase):
    """Unit tests for the {% attr %} template tag."""

    def test_with_value(self):
        """Testing {% attr %} with value"""
        t = Template('{% load djblets_utils %}'
                     '<span{% attr "class" %}\n'
                     '{%  if some_bool %}truthy{% endif %}\n'
                     '{% endattr %}>')

        self.assertEqual(
            t.render(Context({
                'some_bool': True,
            })),
            '<span class="truthy">')

    def test_without_value(self):
        """Testing {% attr %} with no value"""
        t = Template('{% load djblets_utils %}'
                     '<span{% attr "class" %}\n'
                     '{%  if some_bool %}falsy{% endif %}\n'
                     '{% endattr %}>')

        self.assertEqual(
            t.render(Context({
                'some_bool': False,
            })),
            '<span>')

    def test_escapes_value(self):
        """Testing {% attr %} escapes value"""
        t = Template('{% load djblets_utils %}'
                     '<span{% attr "data-foo" %}<hello>{% endattr %}>')

        self.assertEqual(
            t.render(Context()),
            '<span data-foo="&lt;hello&gt;">')

    def test_condenses_whitespace(self):
        """Testing {% attr %} condenses/strips extra whitespace by default"""
        t = Template('{% load djblets_utils %}'
                     '<span{% attr "data-foo" %}\n'
                     'some    \n\n'
                     'value\n'
                     '{% endattr %}>')

        self.assertEqual(
            t.render(Context()),
            '<span data-foo="some value">')

    def test_with_nocondense_preserves_whitespace(self):
        """Testing {% attr %} with "nocondense" option preserves whitespace"""
        t = Template('{% load djblets_utils %}'
                     '<span{% attr "data-foo" nocondense %}\n'
                     'some    \n\n'
                     'value\n'
                     '{% endattr %}>')

        self.assertEqual(
            t.render(Context()),
            '<span data-foo="\nsome    \n\nvalue\n">')


class DefineVarTagTests(TestCase):
    """Unit tests for the {% definevar %} template tag."""

    def test_basic_usage(self):
        """Testing {% definevar %}"""
        t = Template('{% load djblets_utils %}'
                     '{% definevar "myvar" %}\n'
                     'test{{num}}\n'
                     '{% enddefinevar %}'
                     '{{myvar}}')

        self.assertEqual(
            t.render(Context({
                'num': 123,
            })),
            '\ntest123\n')

    def test_with_global(self):
        """Testing {% definevar %} with global option"""
        t = Template(
            '{% load djblets_utils %}'
            '{% block main %}'
            '{%  block inner %}'
            '{%   definevar "myvar" global %}{{num}}{% enddefinevar %}'
            '{%  endblock %}'
            '{% endblock %}'
            '[{{myvar}}]')

        self.assertEqual(
            t.render(Context({
                'num': 123,
            })),
            '[123]')

    def test_with_strip(self):
        """Testing {% definevar %} with strip option"""
        t = Template('{% load djblets_utils %}'
                     '{% definevar "myvar" strip %}\n'
                     '<span>\n'
                     ' <strong>\n'
                     '  test{{num}}\n'
                     ' </strong>\n'
                     '</span>\n'
                     '{% enddefinevar %}'
                     '[{{myvar}}]')

        self.assertEqual(
            t.render(Context({
                'num': 123,
            })),
            '[<span>\n <strong>\n  test123\n </strong>\n</span>]')

    def test_with_spaceless(self):
        """Testing {% definevar %} with spaceless option"""
        t = Template('{% load djblets_utils %}'
                     '{% definevar "myvar" spaceless %}\n'
                     '<span>\n'
                     ' <strong>\n'
                     '  test{{num}}\n'
                     ' </strong>\n'
                     '</span>\n'
                     '{% enddefinevar %}'
                     '[{{myvar}}]')

        self.assertEqual(
            t.render(Context({
                'num': 123,
            })),
            '[<span><strong>\n  test123\n </strong></span>]')

    def test_with_unsafe(self):
        """Testing {% definevar %} with unsafe option"""
        t = Template('{% load djblets_utils %}'
                     '{% definevar "myvar" unsafe %}<hello>{% enddefinevar %}'
                     '{{myvar}}')

        self.assertEqual(t.render(Context()), '&lt;hello&gt;')


class EscapeSpacesFilterTests(TestCase):
    """Unit tests for the {{...|escapespaces}} template filter."""

    def test_with_single_space(self):
        """Testing {{...|escapespaces}} with single space"""
        self.assertEqual(escapespaces('Hi there'),
                         'Hi there')

    def test_with_multiple_spaces(self):
        """Testing {{...|escapespaces}} with multiple consecutive spaces"""
        self.assertEqual(escapespaces('Hi  there'),
                         'Hi&nbsp; there')

    def test_with_newline(self):
        """Testing {{...|escapespaces}} with newline"""
        self.assertEqual(escapespaces('Hi  there\n'),
                         'Hi&nbsp; there<br />')


class GetAttrFilterTests(TestCase):
    """Unit tests for the {{...|getattr:...}} template filter."""

    def test_with_attr_found(self):
        """Testing {{...|getattr:...}} with attribute found"""
        class MyObject(object):
            key = 'value'

        t = Template('{% load djblets_utils %}'
                     '{{obj|getattr:"key"}}')

        self.assertEqual(
            t.render(Context({
                'obj': MyObject(),
            })),
            'value')

    def test_with_attr_not_found(self):
        """Testing {{...|getattr:...}} with attribute not found"""
        class MyObject(object):
            key = 'value'

        t = Template('{% load djblets_utils %}'
                     '{{obj|getattr:"bad_key"}}')

        self.assertEqual(
            t.render(Context({
                'obj': MyObject(),
            })),
            'None')

    def test_with_attr_not_found_and_default_if_none(self):
        """Testing {{...|getattr:...}} with attribute not found and
        default_if_none fallback filter
        """
        class MyObject(object):
            key = 'value'

        t = Template('{% load djblets_utils %}'
                     '{{obj|getattr:"bad_key"|default_if_none:"doh"}}')

        self.assertEqual(
            t.render(Context({
                'obj': MyObject(),
            })),
            'doh')


class HumanizeListFilterTests(TestCase):
    """Unit tests for the {{...|humanize_list}} template filter."""

    def test_with_empty_list(self):
        """Testing {{...|humanize_list}} with empty list"""
        self.assertEqual(humanize_list([]),
                         '')

    def test_with_1_item(self):
        """Testing {{...|humanize_list}} with 1 item"""
        self.assertEqual(humanize_list(['a']),
                         'a')

    def test_with_2_items(self):
        """Testing {{...|humanize_list}} with 2 items"""
        self.assertEqual(humanize_list(['a', 'b']),
                         'a and b')

    def test_with_3_items(self):
        """Testing {{...|humanize_list}} with 3 items"""
        self.assertEqual(humanize_list(['a', 'b', 'c']),
                         'a, b and c')

    def test_with_4_items(self):
        """Testing {{...|humanize_list}} with 4 items"""
        self.assertEqual(humanize_list(['a', 'b', 'c', 'd']),
                         'a, b, c, and d')


class IncludeAsStringTagTests(TestCase):
    """Unit tests for the {% include_as_string %} template tag."""

    def test_basic_usage(self):
        """Testing {% include_as_string %}"""
        t = Template('{% load djblets_utils %}'
                     '{% include_as_string template_name %}')

        self.assertEqual(
            t.render(Context({
                'template_name': 'testing/foo.html',
                'foo': 1,
                'bar': 2,
            })),
            "'1 2\\\n'")


class IndentFilterTests(TestCase):
    """Unit tests for the {{...|indent}} template filter."""

    def test_with_default_indent(self):
        """Testing {{...|indent}} with default indentation level"""
        self.assertEqual(indent('foo'), '    foo')

    def test_with_custom_indent(self):
        """Testing {{...|indent}} with custom indentation level"""
        self.assertEqual(indent('foo', 3), '   foo')

    def test_with_multiple_lines(self):
        """Testing {{...|indent}} with multiple lines"""
        self.assertEqual(indent('foo\nbar'),
                         '    foo\n    bar')


class QuerystringWithTagTests(TestCase):
    """Unit tests for the {% querystring_with %} template tag."""

    def setUp(self):
        super(QuerystringWithTagTests, self).setUp()
        self.request = HttpRequest()
        warnings.simplefilter('always')

    def test_basic_usage(self):
        """Testing {% querystring_with %}"""
        t = Template('{% load djblets_utils %}'
                     '{% querystring_with "foo" "bar" %}')

        with warnings.catch_warnings(record=True) as w:
            t.render(Context({
                'request': self.request,
            }))

        self.assertEqual(len(w), 1)
        deprecation_message = w[0].message

        self.assertIsInstance(deprecation_message, RemovedInDjblets30Warning)
        self.assertEqual(
            six.text_type(deprecation_message),
            '{% querystring_with "foo" "bar" %} is deprecated and will be '
            'removed in a future version of Djblets. Please use '
            '{% querystring "mode" "foo=bar" %} instead.')

        self.assertEqual(
            t.render(Context({
                'request': HttpRequest(),
            })),
            '?foo=bar')

    def test_with_tag_existing_query(self):
        """Testing {% querystring_with %} with an existing query"""
        t = Template('{% load djblets_utils %}'
                     '{% querystring_with "foo" "bar" %}')

        self.request.GET = OrderedDict([
            ('a', '1'),
            ('b', '2'),
        ])

        self.assertEqual(
            t.render(Context({
                'request': self.request,
            })),
            '?a=1&amp;b=2&amp;foo=bar')

    def test_with_existing_query_override(self):
        """Testing {% querystring_with %} with an existing query that gets
        overriden
        """
        t = Template('{% load djblets_utils %}'
                     '{% querystring_with "foo" "bar" %}')

        self.request.GET = {
            'foo': 'foo',
            'bar': 'baz',
        }

        self.assertEqual(
            t.render(Context({
                'request': self.request,
            })), '?bar=baz&amp;foo=bar')


class QuerystringTagTests(TestCase):
    """Unit tests for the {% querystring %} template tag."""

    def test_update_basic_usage(self):
        """Testing {% querystring "update" %} basic usage"""
        self.assertEqual(
            self._render_tag(tag='{% querystring "update" "foo=bar" %}',
                             query_str='foo=bar'),
            '?foo=bar')

    def test_update_with_tag_existing_query(self):
        """Testing {% querystring "update" %} with an existing query"""
        rendered_result = self._render_tag(
            tag='{% querystring "update" "foo=bar" %}',
            query_str='a=1&b=2')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('a=1&;b=2&;foo=bar'))

    def test_update_with_existing_query_override(self):
        """Testing {% querystring "update" %} with an existing query that gets
        overridden
        """
        rendered_result = self._render_tag(
            tag='{% querystring "update" "foo=bar" %}',
            query_str='foo=foo&bar=baz')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('foo=bar&bar=baz'))

    def test_update_with_existing_query_with_two_args_override(self):
        """Testing {% querystring "update" %} with two args that get
        overridden
        """
        rendered_result = self._render_tag(
            tag='{% querystring "update" "foo=bar" "qux=baz" %}',
            query_str='foo=foo&bar=bar&baz=baz&qux=qux')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('foo=bar&bar=bar&baz=baz&qux=baz'))

    def test_update_with_no_value(self):
        """Testing {% querystring "update" %} with no value"""
        rendered_result = self._render_tag(
            tag='{% querystring "update" "foo" %}',
            query_str='')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]), QueryDict('foo='))

    def test_update_with_multiple_values(self):
        """Testing {% querystring "update" %} with multiple values"""
        rendered_result = self._render_tag(
            tag='{% querystring "update" "foo=bar=baz" %}',
            query_str='foo=foo')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('foo=bar=baz'))

    def test_update_with_empty_value(self):
        """Testing {% querystring "update" %} with empty value"""
        rendered_result = self._render_tag(
            tag='{% querystring "update" "foo=" %}',
            query_str='')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]), QueryDict('foo='))

    def test_update_with_no_key(self):
        """Testing {% querystring "update" %} with no key"""
        rendered_result = self._render_tag(
            tag='{% querystring "update" "=foo" %}',
            query_str='')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]), QueryDict('=foo'))

    def test_update_with_querystring_key_overide(self):
        """Testing {% querystring "update" %} by updating multiple values of a
        key value gets overriden
        """
        rendered_result = self._render_tag(
            tag='{% querystring "update" "a=1" "a=2" %}',
            query_str='foo=foo')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('foo=foo&a=2'))

    def test_with_updating_multiple_values_of_a_key(self):
        """Testing {% querystring "update" %} by updating multiple values of a
        key value
        """
        rendered_result = self._render_tag(
            tag='{% querystring "update" "a=1&a=2" %}',
            query_str='foo=foo')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('foo=foo&a=1&a=2'))

    def test_append_with_basic_usage(self):
        """Testing {% querystring "append" %} with appending on to an existing
        key
        """
        rendered_result = self._render_tag(
            tag='{% querystring "append" "foo=baz" %}',
            query_str='foo=foo&bar=bar')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('foo=foo&foo=baz&bar=bar'))

    def test_append_with_multiple_values_and_same_key(self):
        """Testing {% querystring "append" %} with appending multiple values of
        a key
        """
        rendered_result = self._render_tag(
            tag='{% querystring "append" "a=1&a=2&a=3" %}',
            query_str='a=0&&b=2&c=3')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('a=0&a=1&a=2&a=3&b=2&c=3'))

    def test_append_with_multiple_values_and_same_key_seperated(self):
        """Testing {% querystring "append" %} with appending multiple values of
        a key fragment
        """
        rendered_result = self._render_tag(
            tag='{% querystring "append" "a=1" "a=2" "a=3" %}',
            query_str='a=0&&b=2&c=3')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('a=0&a=1&a=2&a=3&b=2&c=3'))

    def test_append_with_new_key(self):
        """Testing {% querystring "append" %} with appending new key-value
        pair
        """
        rendered_result = self._render_tag(
            tag='{% querystring "append" "d=4" %}',
            query_str='a=1&b=2&c=3')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('a=1&b=2&c=3&d=4'))

    def test_remove_with_basic_usage(self):
        """Testing {% querystring "remove" %} by removing a single instance of
        key
        """
        rendered_result = self._render_tag(
            tag='{% querystring "remove" "foo" %}',
            query_str='foo=foo&bar=bar')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]), QueryDict('bar=bar'))

    def test_remove_with_key_not_in_querystring(self):
        """Testing {% querystring "remove" %} by attempting to remove a
        non-existing key
        """
        rendered_result = self._render_tag(
            tag='{% querystring "remove" "baz" %}',
            query_str='foo=foo&bar=bar')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('foo=foo&bar=bar'))

    def test_remove_with_key_appearing_multiple_times(self):
        """Testing {% querystring "remove" %} by removing all instances of a
        key
        """
        rendered_result = self._render_tag(
            tag='{% querystring "remove" "foo" %}',
            query_str='foo=foo&foo=bar&bar=bar')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]), QueryDict('bar=bar'))

    def test_remove_for_specific_key_value_pairs(self):
        """Testing {% querystring "remove" %} by removing a specific key-value
        pair
        """
        rendered_result = self._render_tag(
            tag='{% querystring "remove" "a=4" %}',
            query_str='a=1&a=2&a=3&a=4')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('a=1&a=2&a=3&'))

    def test_remove_with_no_key(self):
        """Testing {% querystring "remove" %} by removing a value with
        no key
        """
        rendered_result = self._render_tag(
            tag='{% querystring "remove" "=foo" %}',
            query_str='foo=foo&foo=bar&baz=baz&=foo')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]),
                         QueryDict('foo=foo&foo=bar&baz=baz'))

    def test_remove_with_no_value(self):
        """Testing {% querystring "remove" %} by removing a key with no
        value
        """
        rendered_result = self._render_tag(
            tag='{% querystring "remove" "foo=" %}',
            query_str='foo=foo&foo=bar&foo=&baz=baz')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]), QueryDict('baz=baz'))

    def test_remove_with_multiple_removes(self):
        """Testing {% querystring "remove" %} by removing multiple keys and
        values
        """
        rendered_result = self._render_tag(
            tag='{% querystring "remove" "foo" "bar" "baz=1" %}',
            query_str='foo=foo&bar=bar&foo=&baz=1&qux=qux')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]), QueryDict('qux=qux'))

    def test_remove_with_multiple_specific_values(self):
        """Testing {% querystring "remove" %} by removing multiple specific
        key-value pairs
        """
        rendered_result = self._render_tag(
            tag='{% querystring "remove" "foo=1" "foo=2" %}',
            query_str='foo=1&foo=2&foo=3')

        self.assertTrue(rendered_result.startswith('?'))
        self.assertEqual(QueryDict(rendered_result[1:]), QueryDict('foo=3'))

    def _render_tag(self, tag, query_str):
        """Returned a rendered template tag using a query string.

        This will render a ``querystring`` template using the provided template
        tag, with auto-escaping turned off, and with the given query string as
        would be provided in a URL.

        Args:
            tag (unicode):
                The template tag line (``{% querystring ... %}``).

            query_str (unicode):
                The query string to simulate being in the URL.

        Returns:
            unicode:
            The rendered content, without escaping.
        """
        t = Template('{%% load djblets_utils %%}'
                     '{%% autoescape off %%}%s{%% endautoescape %%}'
                     % tag)

        request = HttpRequest()

        if query_str:
            request.GET = QueryDict(query_str)

        return t.render(Context({
            'request': request,
        })).replace('&amp;', '&')



"""Field for storing data as Base64-encoded values.

See :py:class:`Base64Field` for the main field type and usage instructions.
"""

from __future__ import unicode_literals

try:
    # Python >= 3.1
    from base64 import (decodebytes as base64_decode,
                        encodebytes as base64_encode)
except ImportError:
    # Python < 3.1
    from base64 import (decodestring as base64_decode,
                        encodestring as base64_encode)

from django.db import models
from django.utils import six
from django.utils.encoding import force_text, smart_text
from django.utils.translation import ugettext as _


class Base64TypeError(TypeError):
    """Error indicating an invalid value type was provided for the field."""

    def __init__(self, value):
        """Initialize the error.

        Args:
            value (object):
                The invalid value that was passed.
        """
        super(Base64TypeError, self).__init__(
            _('The provided value must be a string type or a '
              'Base64DecodedValue, not %r.')
            % type(value))


class Base64DecodedValue(bytes):
    """An identifiable wrapper around byte string values for Base64Field.

    This wraps any values coming from :py:class:`Base64Field`, helping make
    a distinction between arbitrary strings and decoded strings from the
    database. It helps prevent double-encoding or double-decoding of data.
    """


class Base64FieldCreator(object):
    """Property-like class used to store/retrieve Base64 values.

    This works much like a property, and takes care of encoding strings for
    storage and decoding them on retrieval. It's set internally by
    :py:class:`Base64Field` in place of the normal field attribute.
    """

    def __init__(self, field):
        """Initialize the creator.

        Args:
            field (Base64Field):
                The field owning this object.
        """
        self.field = field

    def __set__(self, obj, value):
        """Set a new value in the field.

        If this is a :py:class:`Base64DecodedValue`, or the model is new and
        hasn't yet been persisted to the database, the value will be encoded
        and stored. Otherwise, if it's a standard string value or the model
        is not new, it's assumed that this is encoded data for storage, and
        will be stored directly.

        Args:
            obj (django.db.models.Model):
                The model owning the field.

            value (object):
                The value being set. This must be a valid string value or
                :py:class:`Base64DecodedValue`.

        Raises:
            Base64TypeError:
                The type of value provided could not be set.
        """
        if value is not None:
            if not isinstance(value, (bytes, six.memoryview, six.text_type)):
                raise Base64TypeError(value)

            pk_val = obj._get_pk_val(obj.__class__._meta)
            pk_set = pk_val is not None and smart_text(pk_val) != ''

            if isinstance(value, six.text_type):
                if value == r'\x':
                    # On Djblets 2.0 with Python 3 on Postgres, empty byte
                    # strings were being stored as the literal 2-byte string
                    # "\x". Check for this and convert back.
                    value = b''
                else:
                    value = value.encode('utf-8')
            elif isinstance(value, six.memoryview):
                value = bytes(value)

            if isinstance(value, Base64DecodedValue) or not pk_set:
                value = base64_encode(value)

        obj.__dict__[self.field.name] = value

    def __get__(self, obj, *args, **kwargs):
        """Return a decoded value from the field.

        Args:
            obj (django.db.models.Model):
                The model owning the field.

            *args (tuple):
                Unused positional arguments.

            **kwargs (dict):
                Unused keyword arguments.

        Returns:
            Base64DecodedValue:
            The decoded value from the field. If no value has yet been stored,
            this will return ``None`` instead.

        Raises:
            AttributeError:
            A ``None`` value was passed for the object.
        """
        if obj is None:
            raise AttributeError('Can only be accessed via an instance.')

        value = obj.__dict__[self.field.name]

        if value is not None:
            value = Base64DecodedValue(base64_decode(value))

        return value


class Base64Field(models.TextField):
    """A text field for storing Base64-encoded values.

    This is used to store data (such as binary data or encoding-sensitive data)
    to the database in a Base64 encoding. This is useful if you're dealing
    with unknown encodings and must guarantee that no modifications to the text
    occurs and that you can read/write the data in any database with any
    encoding.

    When accessing this field on an instance of a model, a
    :py:class:`Base64DecodedValue` will be returned consisting of the decoded
    data. This is a byte string, and can be treated as such. If set back into
    the field, it will be re-encoded and stored.

    When writing to the field, the behavior changes based on the type of value
    and the state of the model:

    * If the model instance is new (has not yet been saved in the database),
      any string set will be encoded. This allows the value to be passed during
      a ``create()`` call.

    * If the model is not new, any string that's set will be assumed to be
      encoded by the caller.

    * Passing a :py:class:`Base64DecodedValue` byte string will always cause
      the stored data to be encoded.

    The field also adds a :samp:`get_{fieldname}_base64()` method to the
    class, which returns the raw Base64 encoded content from the database.
    """

    def contribute_to_class(self, cls, name):
        """Set attributes on a new model class.

        This is called when constructing a model class making use of this
        field. It sets the field's attribute to a
        :py:class:`Base64FieldCreator` and adds a
        :samp:`get_{fieldname}_base64()` method to the class.

        Args:
            cls (type):
                The class to add the arguments to.

            name (str):
                The name of the field.
        """
        super(Base64Field, self).contribute_to_class(cls, name)

        setattr(cls, self.name, Base64FieldCreator(self))
        setattr(cls, 'get_%s_base64' % self.name,
                lambda model_instance: model_instance.__dict__[self.name])

    def get_prep_value(self, value):
        """Return a value prepared for the field.

        This prepares the value for use in database operations (saving
        or querying). It will convert the value into a Unicode Base64-encoded
        string.

        Args:
            value (object):
                The value to prepare. This is expected to be a string or a
                :py:class:`Base64DecodedValue`. If the latter, it will be
                encoded.

        Returns:
            unicode:
            The resulting value.

        Raises:
            Base64TypeError:
                The type of value provided could not be prepared for writing.
        """
        if value is not None:
            if isinstance(value, Base64DecodedValue):
                value = base64_encode(value).decode('utf-8')
            elif isinstance(value, bytes):
                value = value.decode('utf-8')
            elif isinstance(value, six.memoryview):
                value = force_text(bytes(value))
            elif not isinstance(value, six.text_type):
                raise Base64TypeError(value)

        return value

    def to_python(self, value):
        """Return a Python representation of a value for the field.

        This will decode the value (if not already decoded) and return it.

        Args:
            value (object):
                The value to return a decoded value for.

        Returns:
            Base64DecodedValue:
            The decoded version of the provided value.

        Raises:
            Base64TypeError:
                The type of value provided could not be prepared for writing.
        """
        if value is not None and not isinstance(value, Base64DecodedValue):
            if isinstance(value, six.text_type):
                value = value.encode('utf-8')
            elif isinstance(value, six.memoryview):
                value = bytes(value)
            elif not isinstance(value, bytes):
                raise Base64TypeError(value)

            value = Base64DecodedValue(base64_decode(value))

        return value

    def value_to_string(self, obj):
        """Return a string representation of the value from a model.

        The returned value will be a Base64-encoded string value.

        Args:
            obj (django.db.models.Model):
                The model instance owning the field and value.

        Returns:
            bytes:
            The Base64-encoded byte string for the stored value.
        """
        value = self.value_from_object(obj)
        assert value is None or isinstance(value, Base64DecodedValue)

        if value is not None:
            value = base64_encode(value)

        return value



"""Compatibility for features removed in Python 3.

Note that this module is not considered API-stable. Changes may be made that
remove or alter functionality in the future. Please use at your own risk.
"""

from __future__ import absolute_import, unicode_literals

from django.utils import six


if six.PY2:
    cmp = cmp
elif six.PY3:
    cmp = lambda x, y: (x > y) - (x < y)


__all__ = (
    'cmp',
)



#
# djblets_utils.py -- Various utility template tags
#
# Copyright (c) 2007-2009  Christian Hammond
# Copyright (c) 2007-2009  David Trowbridge
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from __future__ import unicode_literals

import datetime
import os
import re
import warnings

from django import template
from django.http import QueryDict
from django.template import TemplateSyntaxError, Variable
from django.template.defaultfilters import stringfilter
from django.utils import six
from django.utils.html import escape, format_html, strip_spaces_between_tags
from django.utils.safestring import mark_safe
from django.utils.six.moves import range
from django.utils.six.moves.urllib.parse import urlencode
from django.utils.timezone import is_aware

from djblets.deprecation import RemovedInDjblets30Warning
from djblets.util.compat.django.template.loader import render_to_string
from djblets.util.decorators import blocktag
from djblets.util.dates import get_tz_aware_utcnow
from djblets.util.http import get_url_params_except
from djblets.util.humanize import humanize_list


register = template.Library()


WS_RE = re.compile('\s+')


@register.tag
@blocktag(resolve_vars=False)
def definevar(context, nodelist, varname, *options):
    """Define a variable for later use in the template.

    The variable defined can be used within the same context (such as the
    same block or for loop). This is useful for caching a portion of a
    template that would otherwise be expensive to repeatedly compute.

    Version Added:
        2.0:
        Added the new ``global`` option.

    Version Added:
        1.0:
        Added new ``strip``, ``spaceless``, and ``unsafe`` options.

    Args:
        varname (unicode):
            The variable name.

        *options (list of unicode, optional):
            A list of options passed. This supports the following:

            ``global``:
                Register the variable in the top-level context, for other
                blocks to see.

                Note that the ordering of registration and usage is important,
                so consumers are advised to have a dedicated template block
                for this purpose. Also note that if a later block defines a
                variable with the same name, that will take precedence.

            ``strip``:
                Strip whitespace at the beginning/end of the value.

            ``spaceless``:
                Strip whitespace at the beginning/end and remove all spaces
                between tags. This implies ``strip``.

            ``unsafe``:
                Mark the text as unsafe. The contents will be HTML-escaped when
                inserted into the page.

        block_content (unicode):
            The block content to set in the variable.

    Example:
        .. code-block:: html+django

           {% definevar "myvar1" %}
           {%  expensive_tag %}
           {% enddefinevar %}

           {% definevar "myvar2" spaceless %}
           <div>
            <a href="#">Click me!</a>
           </div>
           {% enddefinevar %}

           {{myvar1}}
           {{myvar2}}
    """
    varname = Variable(varname).resolve(context)
    result = nodelist.render(context)

    if 'spaceless' in options:
        result = strip_spaces_between_tags(result.strip())
    elif 'strip' in options:
        result = result.strip()

    if 'unsafe' in options:
        # Unicode strings are inherently "unsafe".
        result = six.text_type(result)
    else:
        result = mark_safe(result)

    if 'global' in options:
        # Note that we're setting at index 1. That's the first mutable
        # context dictionary. Index 0 is reserved for primitives (True,
        # False, None).
        context.dicts[1][varname] = result
    else:
        context[varname] = result

    return ''


@register.tag
@blocktag
def ifuserorperm(context, nodelist, user, perm):
    """Render content only for a given user or a user with a given permission.

    The content will only be rendered if the logged in user matches the
    specified user, or the logged in user has the specified permission.

    This is useful when you want to restrict some content to the owner of an
    object or to a privileged user that has the abilities of the owner.

    Args:
        user (django.contrib.auth.models.User):
            The user to limit access to, unless the logged in user has the
            specified permission.

        perm (unicode):
            The permission to require, if the logged in user does not match
            the specified user.

        block_content (unicode):
            The block content to render.

    Returns:
        The content, if the user or permission matches. Otherwise, no content
        will be returned.

    Example:
        .. code-block:: html+django

           {% ifuserorperm myobject.user "myobject.can_change_status" %}
           Owner-specific content here...
           {% endifuserorperm %}
    """
    if _check_userorperm(context, user, perm):
        return nodelist.render(context)

    return ''


@register.tag
@blocktag
def ifnotuserandperm(context, nodelist, user, perm):
    """Render content if a user and permission don't match the logged in user.

    This is the opposite of :py:func:`{% ifuserorperm %} <ifuserorperm>`. It
    will only render content if the logged in user is not the specified user
    and doesn't have the specified permission.

    Args:
        user (django.contrib.auth.models.User):
            The user who cannot see the provided content.

        perm (unicode):
            Any user with this permission will not see the provided content.

        block_content (unicode):
            The block content to render.

    Returns:
        The content, if neither the user nor permission matches. Otherwise, no
        content will be returned.

    Example:
        .. code-block:: html+django

           {% ifuserorperm myobject.user "myobject.can_change_status" %}
           Owner-specific content here...
           {% endifuserorperm %}

           {% ifnotuserandperm myobject.user "myobject.can_change_status" %}
           Another owner-specific content here...
           {% endifnotuserandperm %}
    """
    if not _check_userorperm(context, user, perm):
        return nodelist.render(context)

    return ''


def _check_userorperm(context, user, perm):
    from django.contrib.auth.models import AnonymousUser, User

    req_user = context.get('user', None)

    if isinstance(req_user, AnonymousUser):
        return False

    if req_user.has_perm(perm):
        return True

    return ((isinstance(user, User) and user == req_user) or
            user == req_user.pk)


@register.simple_tag(takes_context=True)
def include_as_string(context, template_name):
    """Include the contents of a template as an escaped string.

    This is primarily for use with JavaScript. It allows another template
    to be rendered (with the current context) and returned as an escaped
    string.

    Args:
        template_name (unicode):
            The name of the template to render.

    Returns:
        The escaped content from the template.

    Example:
        .. code-block:: html+django

           <script>
           var s = {% include_as_string "message.txt" %};
           </script>
    """
    s = render_to_string(template_name, context)
    s = s.replace("'", "\\'")
    s = s.replace("\n", "\\\n")

    # Since this works like {% include %}, we have to trust the resulting
    # content here. It's still possible that a nefarious template could cause
    # problems, but this is the responsibility of the caller.
    #
    # In prior versions of Django (< 1.9), this was implicitly marked safe.
    return mark_safe("'%s'" % s)


@register.tag
@blocktag(resolve_vars=False)
def attr(context, nodelist, attrname, *options):
    """Set an HTML attribute to a value if the value is not an empty string.

    This is a handy way of adding attributes with non-empty values to an
    HTML element without requiring several `{% if %}` tags.

    The contents will be stripped and all whitespace within condensed before
    being considered or rendered. This can be turned off (restoring pre-1.0
    behavior) by passing ``nocondense`` as an option.

    .. versionchanged:: 1.0

       Prior to this release, all whitespace before/after/within the
       attribute value was preserved. Now ``nocondense`` is required for this
       behavior.

       The value is now escaped as well. Previously the value was assumed to
       be safe, requiring the consumer to escape the contents.

    Args:
        attrname (unicode):
            The name for the HTML attribute.

        *options (list unicode, optional):
            A list of options passed. This supports the following:

            ``nocondense``:
                Preserves all whitespace in the value.

        block_content (unicode):
            The block content to render for the attribute value.

    Returns:
        An attribute in the form of ``key="value"``, if the value (the
        block content) is not empty.

    Example:
        .. code-block:: html+django

           <div{% attr "class" %}{{obj.name}}{% endattr %}>
           <div{% attr "data-description" nocondense %}
               Space-sensitive

               whitespace
           {% endattr %}>
    """
    attrname = Variable(attrname).resolve(context)
    content = nodelist.render(context)

    if 'nocondense' not in options:
        content = WS_RE.sub(' ', content.strip())

    if not content:
        return ''

    return format_html(' {0}="{1}"', attrname, six.text_type(content))


@register.filter
def escapespaces(value):
    """HTML-escape all spaces with ``&nbsp;`` and newlines with ``<br />``.

    Args:
        value (unicode):
            The value to escape.

    Returns:
        unicode:
        The same text, but escaped.

    Example:
        .. code-block:: html+django

           <div class="text">
            {{obj.description|escapespaces}}
           </div>
    """
    return value.replace('  ', '&nbsp; ').replace('\n', '<br />')


@register.simple_tag
def ageid(timestamp):
    """Return an ID based on the difference between now and a timestamp.

    This can be used to help show the age of an item in days. It will
    generate an ID in the form of :samp:`age{num}` ranging from
    ``age1`` (today) through ``age5`` (4+ days old).

    This is a specialty function, and is not expected to be useful in many
    cases.

    Args:
        timestamp (datetime.datetime):
            The timestamp to compare to.

    Returns:
        unicode:
        The ID. One of ``age1``, ``age2``, ``age3``, ``age4``, or ``age5``.

    Example:
        .. code-block:: html+django

           <div class="{% ageid obj.timestamp %}">
            {{obj.timestamp}}
           </div>
    """
    if timestamp is None:
        return ""

    # Convert datetime.date into datetime.datetime
    if timestamp.__class__ is not datetime.datetime:
        timestamp = datetime.datetime(timestamp.year, timestamp.month,
                                      timestamp.day)

    now = datetime.datetime.utcnow()

    if is_aware(timestamp):
        now = get_tz_aware_utcnow()

    delta = now - (timestamp -
                   datetime.timedelta(0, 0, timestamp.microsecond))

    if delta.days == 0:
        return "age1"
    elif delta.days == 1:
        return "age2"
    elif delta.days == 2:
        return "age3"
    elif delta.days == 3:
        return "age4"
    else:
        return "age5"


@register.filter
def user_displayname(user):
    """Return the display name of a user.

    If the user has a full name set, it will be returned. Otherwise, the
    username will be returned.

    Args:
        user (django.contrib.auth.models.User):
            The user whose full name or username will be returned.

    Returns:
        unicode:
        The full name of the user, if set, or the username as a fallback.

    Example:
        .. code-block:: html+django

           Welcome, {{user|user_displayname}}!
    """
    return user.get_full_name() or user.username


register.filter('humanize_list', humanize_list)


@register.filter
def contains(container, value):
    """Return whether the specified value is in the specified container.

    This is equivalent to a ``if value in container`` statement in Python.

    Args:
        container (object):
            The list, dictionary, or other object that may or may not
            contain the value.

        value (object):
            The value being checked.

    Returns:
        bool:
        ``True`` if the value is in the container. Otherwise, ``False``.

    Example:
        .. code-block:: html+django

           {% if usernames|contains:"bob" %}
             Hi, Bob!
           {% endif %}
    """
    return value in container


@register.filter('getattr')
def getattr_filter(obj, name):
    """Return the attribute of a specified name from an object.

    This is equivalent to a ``getattr(obj, name, None)`` statement in
    Python.

    Version Added:
        2.0

    Args:
        obj (object):
            The object that contains the attribute.

        name (unicode):
            The name of the key to look up in the container.

    Returns:
        object:
        The attribute value.

    Example:
        .. code-block:: html+django

           {% for name in attrs %}
             {{name}}: {{obj|getattr:name}}
           {% endfor %}

           {{obj|getattr:other_attr|default_if_none:"my fallback"}}
    """
    return getattr(obj, name, None)


@register.filter
def getitem(container, key):
    """Return the value of a specified key from a container.

    This is equivalent to a ``container[key]`` statement in Python. The
    container must support this operator.

    Args:
        container (object):
            The list, dictionary, or other object that can contain items.

        key (object):
            The key to look up in the container.

    Returns:
        object: The content within the container.

    Example:
        .. code-block:: html+django

           {% for key in keys %}
             {{key}}: {{obj|getitem:key}}
           {% endfor %}
    """
    return container[key]


@register.filter
def exclude_item(container, item):
    """Return a list with the given item excluded.

    Args:
        container (list):
            The list the item will be excluded from.

        item (object):
            The item to exclude from the list.

    Returns:
        list: The list with the item excluded.

    Example:
        .. code-block:: html+django

           {% for item in mylist|exclude_item:"special" %}
             ...
           {% endfor %}
    """
    if not isinstance(container, list):
        raise TemplateSyntaxError("remove_item expects a list")

    container = list(container)

    try:
        container.remove(item)
    except ValueError:
        pass

    return container


@register.filter
def indent(value, numspaces=4):
    """Indent a string by the specified number of spaces.

    This is especially useful for preformatted content.

    Args:
        value (unicode):
            The value containing text to indent.

        numspaces (int, optional):
            The number of spaces to indent the text. Defaults to 4 spaces.

    Returns:
        unicode: The indented text.

    Example:
        .. code-block:: html+django

           <pre>
           The traceback was:

           {{traceback|indent:2}}
           </pre>
    """
    indent_str = ' ' * numspaces
    return indent_str + value.replace('\n', '\n' + indent_str)


@register.filter
def basename(path):
    """Return the base name of a path.

    This will be computed based on the path rules from the system the
    server is running on.

    Args:
        path (unicode):
            The path for which to retrieve the base name.

    Returns:
        unicode:
        The base name of the path.

    Example:
        .. code-block:: html+django

           The file is contained within <tt>{{path|basename}}</tt>.
    """
    return os.path.basename(path)


@register.filter(name="range")
def range_filter(value):
    """Turn an integer into a range of numbers.

    This is useful for iterating with the "for" tag.

    Args:
        value (int):
            The number of values in the range.

    Returns:
        list:
        The list of numbers in the range.

    Example:
        .. code-block:: html+django

            {% for i in 10|range %}
             {{i}}
            {% endfor %}
    """
    return range(value)


@register.filter
def realname(user):
    """Return the real name of a user, if available, or the username.

    If the user has a full name set, this will return the full name.
    Otherwise, this returns the username.

    .. deprecated:: 0.9

       This is deprecated in favor of :py:func:`user_displayname`.

    Args:
        user (django.contrib.auth.models.User):
            The user whose full name or username will be returned.

    Returns:
        unicode:
        The full name of the user, if set, or the username as a fallback.

    Example:
        .. code-block:: html+django

           Welcome, {{user|realname}}!
    """
    full_name = user.get_full_name()
    if full_name == '':
        return user.username
    else:
        return full_name


@register.filter
@stringfilter
def startswith(s, prefix):
    """Return whether a value starts with another value.

    Args:
        s (unicode):
            The string to check.

        prefix (unicode):
            The prefix to check for.

    Returns:
        bool:
        ``True`` if the string starts with the given prefix.

    Example:
        .. code-block:: html+django

           {% if key|startswith:"__" %}
             ..
           {% endif %}
    """
    return s.startswith(prefix)


@register.filter
@stringfilter
def endswith(s, suffix):
    """Return whether a value ends with another value.

    Args:
        s (unicode):
            The string to check.

        suffix (unicode):
            The suffix to check for.

    Returns:
        bool:
        ``True`` if the string ends with the given suffix.

    Example:
        .. code-block:: html+django

           {% if filename|endswith:".json" %}
             ..
           {% endif %}
    """
    return s.endswith(suffix)


@register.filter
@stringfilter
def paragraphs(text):
    """Return an HTML paragraph for each line of text.

    This iterates through the lines of text given and wraps each in a
    ``<p>`` tag.

    This expects that each paragraph in the string will be on its own line.
    Blank lines are filtered out.

    The text is expected to be HTML-safe already.

    Args:
        text (unicode):
            The text containing at least one line of content.

    Returns:
        unicode:
        The resulting HTML output.

    Example:
        .. code-block:: html+django

           <article>
            {{description|paragraphs}}
           </article>
    """
    s = ""

    for line in text.splitlines():
        if line:
            s += "<p>%s</p>\n" % line

    return mark_safe(s)
paragraphs.is_safe = True


@register.filter
@stringfilter
def split(s, delim=','):
    """Split a string into a list.

    The string can be split by any specified delimiter, and defaults to a
    comma.

    Args:
        s (unicode):
            The string to split.

        delim (unicode, optional):
            The delimiter to split by. Defaults to ``,``.

    Returns:
        list:
        The resulting list of tokens from the string.

    Example:
        .. code-block:: html+django

           {% for token in str|split:'\\t' %}
             ..
           {% endfor %}
    """
    return s.split(delim)


@register.simple_tag(takes_context=True)
def querystring_with(context, attr, value):
    """Return the current page URL with a new query string argument added.

    This makes it easy to add to or replace part of a query string for the
    current page's URL, which may already contain a query string.

    If the page URL already has a query string, a new item is added in the form
    of ``&attr=value``. If it doesn't have a query string, this will start a
    new one in the form of ``?attr=value``.

    If the attribute already exists in the query string, its value will be
    replaced.

    Args:
        context (django.template.context.RequestContext):
            The Django template rendering context.

        attr (unicode):
            The name of the attribute for the new query string argument.

        value (unicode):
            The value of the attribute for the new query string argument.

    Returns:
        unicode:
        The new URL with the modified query string.

    Example:
        .. code-block:: html+django

           <a href="{% querystring_with "sorted" "1" %}">Sort</a>
    """
    warnings.warn(
        '{%% querystring_with "%(attr)s" "%(value)s" %%} is deprecated and '
        'will be removed in a future version of Djblets. Please use '
        '{%% querystring "mode" "%(attr)s=%(value)s" %%} instead.'
        % {
            'attr': attr,
            'value': value,
        },
        RemovedInDjblets30Warning)

    existing_query = get_url_params_except(context['request'].GET, attr)
    new_query = urlencode({attr.encode('utf-8'): value.encode('utf-8')})

    if existing_query:
        result = '?%s&%s' % (existing_query, new_query)
    else:
        result = '?%s' % new_query

    return escape(result)


@register.simple_tag(takes_context=True)
def querystring(context, mode, *args):
    """Return current page URL with new query string containing multiple
    parameters.

    Args:
        context (django.template.context.RequestContext):
            The Django template rendering context.

        mode (unicode):
            How the querystring will be modified. This should be one of the
            following values:

            ``'update'``:
                Replace the values for the specified key(s) in the query
                string.

            ``'append'``:
                Add new values for the specified key(s) to the query string.

            ``'remove'``:
                Remove the specified key(s) from the query string.

                If no value is provided, all instances of the key will be
                removed.

        *args (tuple):
            Multiple querystring fragments (e.g., ``foo=1``) that will be used
            to update the initial querystring.

    Returns:
        unicode:
        The new URL with the modified query string.

    Example:
        .. code-block:: html+django

           <a href="{% querystring "update" 'a=1' 'b=2' %}">Sort</a>

           <a href="{% querystring "append" 'a=1' 'b=2' %}">Sort</a>

           <a href="{% querystring "append" 'a=1&a=2' %}">Sort</a>

           <a href="{% querystring "remove" 'a' %}">Sort</a>
    """
    query = QueryDict('', mutable=True)
    query.update(context['request'].GET)

    if mode == 'update':
        for arg in args:
            parsed = QueryDict(arg)

            for attr in parsed:
                query.setlist(attr, parsed.getlist(attr))
    elif mode == 'remove':
        for arg in args:
            parsed = QueryDict(arg)

            for attr in parsed:
                to_remove = parsed.getlist(attr)

                if to_remove == ['']:
                    query.pop(attr, None)
                else:
                    values = query.getlist(attr)

                    for value in to_remove:
                        try:
                            values.remove(value)
                        except ValueError:
                            pass

                    query.setlist(attr, values)
    elif mode == 'append':
        for arg in args:
            query.update(QueryDict(arg))
    else:
        raise TemplateSyntaxError('Invalid mode for {%% querystring %%}: %s'
                                  % mode)

    return escape('?%s' % query.urlencode())



"""Unit tests for djblets.privacy.consent.forms.MultiConsentRequirementsWidget.
"""

from __future__ import unicode_literals

from djblets.privacy.consent import BaseConsentRequirement
from djblets.privacy.consent.forms import MultiConsentRequirementsWidget
from djblets.testing.testcases import TestCase


class MyConsentRequirement1(BaseConsentRequirement):
    requirement_id = 'my-requirement-1'
    name = 'My Requirement 1'
    summary = 'We would like to use this thing'
    intent_description = 'We need this for testing.'
    data_use_description = 'Sending all the things.'
    icons = {
        '1x': '/static/consent.png',
        '2x': '/static/consent@2x.png',
    }


class MyConsentRequirement2(BaseConsentRequirement):
    requirement_id = 'my-requirement-2'
    name = 'My Requirement 2'
    summary = 'We would also like this'
    intent_description = 'We need this for dancing.'
    data_use_description = 'Dancing all the things.'


class MultiConsentRequirementsWidgetTests(TestCase):
    """Unit tests for MultiConsentRequirementsWidget."""

    def setUp(self):
        super(MultiConsentRequirementsWidgetTests, self).setUp()

        self.widget = MultiConsentRequirementsWidget(consent_requirements=[
            MyConsentRequirement1(),
            MyConsentRequirement2(),
        ])

    def test_render(self):
        """Testing MultiConsentRequirementsWidget.render with value=[]"""
        html = self.widget.render(
            name='consent',
            value=[],
            attrs={
                'id': 'my_consent',
            })

        # Check for the first sub-widget.
        self.assertIn('<div id="my_consent_my-requirement-1"'
                      ' class="privacy-consent-field'
                      ' privacy-consent-field-has-icon">',
                      html)
        self.assertIn("<h2>We would like to use this thing</h2>", html)
        self.assertIn('<p>We need this for testing.</p>', html)
        self.assertIn('<p>Sending all the things.</p>', html)
        self.assertInHTML('<input type="radio"'
                          ' id="my_consent_my-requirement-1_choice_allow"'
                          ' name="consent_my-requirement-1_choice"'
                          ' value="allow">',
                          html)
        self.assertInHTML('<input type="radio"'
                          ' id="my_consent_my-requirement-1_choice_block"'
                          ' name="consent_my-requirement-1_choice"'
                          ' value="block">',
                          html)

        # Check for the second sub-widget.
        self.assertIn('<div id="my_consent_my-requirement-2"'
                      ' class="privacy-consent-field">',
                      html)
        self.assertIn("<h2>We would also like this</h2>", html)
        self.assertIn('<p>We need this for dancing.</p>', html)
        self.assertIn('<p>Dancing all the things.</p>', html)
        self.assertInHTML('<input type="radio"'
                          ' id="my_consent_my-requirement-2_choice_allow"'
                          ' name="consent_my-requirement-2_choice"'
                          ' value="allow">',
                          html)
        self.assertInHTML('<input type="radio"'
                          ' id="my_consent_my-requirement-2_choice_block"'
                          ' name="consent_my-requirement-2_choice"'
                          ' value="block">',
                          html)

    def test_render_with_values(self):
        """Testing MultiConsentRequirementsWidget.render with values"""
        html = self.widget.render(
            name='consent',
            value=['allow', 'block'],
            attrs={
                'id': 'my_consent',
            })

        # Check for the first sub-widget.
        self.assertIn('<div id="my_consent_my-requirement-1"'
                      ' class="privacy-consent-field'
                      ' privacy-consent-field-has-icon">',
                      html)
        self.assertIn("<h2>We would like to use this thing</h2>", html)
        self.assertIn('<p>We need this for testing.</p>', html)
        self.assertIn('<p>Sending all the things.</p>', html)
        self.assertInHTML('<input type="radio"'
                          ' id="my_consent_my-requirement-1_choice_allow"'
                          ' name="consent_my-requirement-1_choice"'
                          ' value="allow" checked>',
                          html)
        self.assertInHTML('<input type="radio"'
                          ' id="my_consent_my-requirement-1_choice_block"'
                          ' name="consent_my-requirement-1_choice"'
                          ' value="block">',
                          html)

        # Check for the second sub-widget.
        self.assertIn('<div id="my_consent_my-requirement-2"'
                      ' class="privacy-consent-field">',
                      html)
        self.assertIn("<h2>We would also like this</h2>", html)
        self.assertIn('<p>We need this for dancing.</p>', html)
        self.assertIn('<p>Dancing all the things.</p>', html)
        self.assertInHTML('<input type="radio"'
                          ' id="my_consent_my-requirement-2_choice_allow"'
                          ' name="consent_my-requirement-2_choice"'
                          ' value="allow">',
                          html)
        self.assertInHTML('<input type="radio"'
                          ' id="my_consent_my-requirement-2_choice_block"'
                          ' name="consent_my-requirement-2_choice"'
                          ' value="block" checked>',
                          html)

    def test_value_from_datadict_with_value(self):
        """Testing MultiConsentRequirementsWidget.value_from_datadict with
        values
        """
        data = {
            'consent_my-requirement-1_choice': 'allow',
            'consent_my-requirement-2_choice': 'block',
        }

        self.assertEqual(
            self.widget.value_from_datadict(data=data,
                                            files={},
                                            name='consent'),
            ['allow', 'block'])

    def test_value_from_datadict_without_values(self):
        """Testing MultiConsentRequirementsWidget.value_from_datadict without
        values
        """
        self.assertEqual(
            self.widget.value_from_datadict(data={}, files={}, name='consent'),
            [None, None])



# TODO: Remove this file once we no longer support a verison of Django
#       prior to 1.7.
#
"""Compatibility functions for working with template contexts."""

from __future__ import unicode_literals

from django.template.context import Context


def flatten_context(context):
    """Flatten a template context into a dictionary.

    Django 1.7 introduced :py:meth:`Context.flatten
    <django.template.Context.flatten>`, which converts a template context into
    a dictionary. However, this doesn't exist on Django 1.6. This compatibility
    function ensures equivalent functionality on both.
    """
    assert isinstance(context, Context)

    if hasattr(context, 'flatten'):
        # Django >= 1.7
        return context.flatten()
    else:
        # Django == 1.6
        new_dict = {}

        for d in context.dicts:
            new_dict.update(d)

        return new_dict



"""Unit tests for djblets.extensions.settings.ExtensionSettings."""

from __future__ import unicode_literals

from mock import Mock

from djblets.extensions.settings import ExtensionSettings
from djblets.extensions.signals import settings_saved
from djblets.testing.testcases import TestCase


class ExtensionSettingsTest(TestCase):
    """Unit tests for djblets.extensions.settings.ExtensionSettings."""

    def setUp(self):
        # Build up a mocked extension
        self.extension = Mock()
        self.extension.registration = Mock()
        self.test_dict = {
            'test_key1': 'test_value1',
            'test_key2': 'test_value2',
        }
        self.extension.registration.settings = self.test_dict
        self.settings = ExtensionSettings(self.extension)

    def test_constructor(self):
        """Testing ExtensionSettings construction"""
        self.assertEqual(self.extension, self.settings.extension)

        # Ensure that the registration settings dict got added.
        self.assertEqual(self.test_dict['test_key1'],
                         self.settings['test_key1'])

    def test_load_updates_dict(self):
        """Testing ExtensionSettings.load correctly updates core dict"""
        new_dict = {
            'test_new_key': 'test_new_value',
            'test_key1': 'new_value',
        }
        self.extension.registration.settings = new_dict
        self.settings.load()

        # Should have added test_new_key, and modified test_key1.
        self.assertEqual(new_dict['test_new_key'],
                         self.settings['test_new_key'])
        self.assertEqual(new_dict['test_key1'], self.settings['test_key1'])

        # Should have left test_key2 alone.
        self.assertEqual(self.test_dict['test_key2'],
                         self.settings['test_key2'])

    def test_load_silently_discards(self):
        """Testing ExtensionSettings.load silently ignores invalid settings"""
        some_string = 'This is a string'
        self.extension.registration.settings = some_string

        try:
            self.settings.load()
        except Exception:
            self.fail("Shouldn't have raised an exception")

    def test_save_updates_database(self):
        """Testing ExtensionSettings.save will correctly update registration"""
        registration = self.extension.registration
        self.settings['test_new_key'] = 'Test new value'
        generated_dict = dict(self.settings)
        self.settings.save()

        self.assertTrue(registration.save.called)
        self.assertEqual(generated_dict, registration.settings)

    def test_save_emits_settings_saved_signal(self):
        """Testing ExtensionSettings.save emits the settings_saved signal"""
        saw = {}

        def on_settings_saved(*args, **kwargs):
            saw['signal'] = True

        settings_saved.connect(on_settings_saved, sender=self.extension)

        self.settings['test_new_key'] = 'Test new value'
        self.settings.save()

        self.assertIn('signal', saw)



from __future__ import unicode_literals

from djblets.features import Feature, FeatureLevel, get_features_registry
from djblets.features.testing import (override_feature_check,
                                      override_feature_checks)
from djblets.testing.testcases import TestCase


class StableFeature(Feature):
    feature_id = 'djblets.tests.stable-feature'
    level = FeatureLevel.STABLE


class ExperimentalFeature(Feature):
    feature_id = 'djblets.tests.experimental-feature'
    level = FeatureLevel.EXPERIMENTAL


class FeatureTestingTests(TestCase):
    """Unit tests for djblets.features.testing."""

    def tearDown(self):
        super(FeatureTestingTests, self).tearDown()

        get_features_registry().reset()

    def test_override_feature_check(self):
        """Testing override_feature_check"""
        stable_feature = StableFeature()
        experimental_feature = ExperimentalFeature()

        self.assertTrue(stable_feature.is_enabled())
        self.assertFalse(experimental_feature.is_enabled())

        with override_feature_check(stable_feature.feature_id, enabled=False):
            self.assertFalse(stable_feature.is_enabled())
            self.assertFalse(experimental_feature.is_enabled())

        with override_feature_check(experimental_feature.feature_id,
                                    enabled=True):
            self.assertTrue(stable_feature.is_enabled())
            self.assertTrue(experimental_feature.is_enabled())

        # Make sure they haven't been permanently overridden.
        self.assertTrue(stable_feature.is_enabled())
        self.assertFalse(experimental_feature.is_enabled())

    def test_override_feature_checks(self):
        """Testing override_feature_checks"""
        stable_feature = StableFeature()
        experimental_feature = ExperimentalFeature()

        self.assertTrue(stable_feature.is_enabled())
        self.assertFalse(experimental_feature.is_enabled())

        feature_states = {
            stable_feature.feature_id: False,
            experimental_feature.feature_id: True,
        }

        with override_feature_checks(feature_states):
            self.assertFalse(stable_feature.is_enabled())
            self.assertTrue(experimental_feature.is_enabled())

        # Make sure they haven't been permanently overridden.
        self.assertTrue(stable_feature.is_enabled())
        self.assertFalse(experimental_feature.is_enabled())



"""Unit tests for djblets.extensions.extension.Extension."""

from __future__ import unicode_literals

from django.core.exceptions import ImproperlyConfigured
from django.utils import six
from kgb import SpyAgency

from djblets.extensions.extension import Extension
from djblets.extensions.hooks import ExtensionHook, ExtensionHookPoint
from djblets.extensions.settings import ExtensionSettings
from djblets.extensions.tests.base import ExtensionTestsMixin
from djblets.testing.testcases import TestCase


@six.add_metaclass(ExtensionHookPoint)
class DummyHook(ExtensionHook):
    def initialize(self):
        self.foo = [1]

    def shutdown(self):
        self.foo.pop()


class ExtensionTest(SpyAgency, ExtensionTestsMixin, TestCase):
    """Unit tests for djblets.extensions.extension.Extension."""

    def test_extension_constructor(self):
        """Testing Extension construction"""
        class TestExtension(Extension):
            pass

        extension = self.setup_extension(TestExtension)

        self.assertEqual(type(extension.settings), ExtensionSettings)
        self.assertEqual(extension, extension.settings.extension)

    def test_shutdown(self):
        """Testing Extension.shutdown"""
        class TestExtension(Extension):
            def initialize(self):
                for index in range(0, 5):
                    DummyHook(self)

        extension = self.setup_extension(TestExtension)

        for hook in extension.hooks:
            self.spy_on(hook.shutdown)

        extension.shutdown()

        for hook in extension.hooks:
            self.assertTrue(hook.shutdown.called)

    def test_shutdown_twice(self):
        """Testing Extension.shutdown when called twice"""
        class TestExtension(Extension):
            def initialize(self):
                for index in range(0, 5):
                    DummyHook(self)

        extension = self.setup_extension(TestExtension)

        for hook in extension.hooks:
            self.spy_on(hook.shutdown)

        extension.shutdown()

        for hook in extension.hooks:
            self.assertTrue(hook.shutdown.called)
            hook.shutdown.reset_calls()

        extension.shutdown()

        for hook in extension.hooks:
            self.assertFalse(hook.shutdown.called)

    def test_admin_urlconf(self):
        """Testing Extension with admin URLConfs"""
        class TestExtension(Extension):
            __module__ = 'djblets.extensions.tests.apps.test'

        extension = self.setup_extension(TestExtension, enable=False)

        try:
            extension.admin_urlconf
        except ImproperlyConfigured:
            self.fail('Should have loaded admin_urls.py')



"""Unit tests for djblets.extensions.hooks.URLHook."""

from __future__ import unicode_literals

from django.conf.urls import include, url

from djblets.extensions.extension import Extension
from djblets.extensions.hooks import URLHook
from djblets.extensions.tests.base import ExtensionTestsMixin
from djblets.testing.testcases import TestCase


class TestExtension(Extension):
    def initialize(self):
        self.patterns = [
            url(r'^url_hook_test/',
                include('djblets.extensions.tests.apps.urls')),
        ]
        self.url_hook = URLHook(self, self.patterns)


class URLHookTest(ExtensionTestsMixin, TestCase):
    """Unit tests for djblets.extensions.hooks.URLHook."""

    def setUp(self):
        super(URLHookTest, self).setUp()

        self.extension = self.setup_extension(TestExtension)

    def test_url_registration(self):
        """Testing URLHook URL registration"""
        self.assertTrue(
            set(self.extension.patterns)
            .issubset(set(self.extension.url_hook.dynamic_urls.url_patterns)))

        # And the URLHook should be added to the extension's list of hooks
        self.assertIn(self.extension.url_hook, self.extension.hooks)

    def test_shutdown_removes_urls(self):
        """Testing URLHook.shutdown"""
        # On shutdown, a URLHook's patterns should no longer be in its
        # parent URL resolver's pattern collection.
        self.extension.url_hook.disable_hook()
        self.assertFalse(
            set(self.extension.patterns).issubset(
                set(self.extension.url_hook.dynamic_urls.url_patterns)))

        # But the URLHook should still be in the extension's list of hooks
        self.assertIn(self.extension.url_hook, self.extension.hooks)



"""Compatibility functions for django.shortcuts.

Note that this module is not considered API-stable. Changes may be made that
remove or alter functionality in the future. Please use at your own risk.
"""

from __future__ import unicode_literals

from django.http import HttpResponse

from djblets.util.compat.django.template.loader import render_to_string


# TODO: Remove render() once we no longer support a version of Django prior
#       to 1.10.

def render(request, template_name, context=None, content_type=None,
           status=None, using=None):
    """Render a template name to a response.

    This provides compatibility for the :py:func:`django.shortcuts.loader`
    across Django 1.6 on up.

    Args:
        request (django.http.HttpRequest):
            The HTTP request from the client.

        template_name (unicode):
            The name of the template to render.

        context (dict, optional):
            The template context.

        content_type (unicode, optional);
            An optional content type to set for the response.

        status (int, optional):
            An optional HTTP status code to set for the response.

        using (unicode, optional):
            The optional template render backend to use to render this
            template. This is ignored for Django 1.6.

    Returns:
        django.http.HttpResponse:
        The resulting HTTP response.
    """
    content = render_to_string(template_name=template_name,
                               context=context,
                               request=request,
                               using=using)

    return HttpResponse(content,
                        content_type=content_type,
                        status=status)



"""A flexible resource for the root of your API resource tree."""

from __future__ import unicode_literals

from collections import namedtuple

from django.conf.urls import url
from django.http import HttpResponseNotModified
from django.utils import six

from djblets.urls.patterns import never_cache_patterns
from djblets.webapi.errors import DOES_NOT_EXIST
from djblets.webapi.resources.base import WebAPIResource
from djblets.webapi.responses import WebAPIResponseError


class RootResource(WebAPIResource):
    """The root of a resource tree.

    This is meant to be instantiated with a list of immediate child
    resources. The result of :py:meth:`get_url_patterns` should be included in
    a project's :file:`urls.py`.
    """

    name = 'root'
    singleton = True

    #: A resource entry returned from :py:meth:`RootResource.walk_resources`.
    #:
    #: Attributes:
    #:      name (unicode):
    #:          The name of current resource being explored.
    #:
    #       list_href (unicode):
    #:          The list_href associated with this resource.
    #:
    #       resource (djblets.webapi.resources.base.WebAPIResource):
    #:          The resource object being explored.
    #:
    #       is_list (bool):
    #:          Whether or not the resource is a list resource.
    ResourceEntry = namedtuple('ResourceEntry',
                               ('name', 'list_href', 'resource', 'is_list'))

    def __init__(self, child_resources=[], include_uri_templates=True):
        super(RootResource, self).__init__()
        self.list_child_resources = child_resources
        self._uri_templates = {}
        self._registered_uri_templates = {}
        self._include_uri_templates = include_uri_templates

    def get_etag(self, request, obj, *args, **kwargs):
        return self.encode_etag(request, repr(obj))

    def get(self, request, *args, **kwargs):
        """Retrieve the list of top-level resources and URL templates."""
        data = self.serialize_root(request, *args, **kwargs)
        etag = self.get_etag(request, data)

        if self.are_cache_headers_current(request, etag=etag):
            return HttpResponseNotModified()

        return 200, data, {
            'ETag': etag,
        }

    def serialize_root(self, request, *args, **kwargs):
        """Serialize the contents of the root resource.

        By default, this just provides links and URI templates. Subclasses
        can override this to provide additional data, or to otherwise
        change the structure of the root resource.
        """
        data = {
            'links': self.get_links(self.list_child_resources,
                                    request=request, *args, **kwargs),
        }

        if self._include_uri_templates:
            data['uri_templates'] = self.get_uri_templates(request, *args,
                                                           **kwargs)

        return data

    def get_uri_templates(self, request, *args, **kwargs):
        """Return all URI templates in the resource tree.

        REST APIs can be very chatty if a client wants to be well-behaved
        and crawl the resource tree asking for the links, instead of
        hard-coding the paths. The benefit is that they can keep from
        breaking when paths change. The downside is that it can take many
        HTTP requests to get the right resource.

        This list of all URI templates allows clients who know the resource
        name and the data they care about to simply plug them into the
        URI template instead of trying to crawl over the whole tree. This
        can make things far more efficient.
        """
        if not self._uri_templates:
            self._uri_templates = {}

        base_href = request.build_absolute_uri()

        if base_href not in self._uri_templates:
            templates = {}
            unassigned_templates = self._registered_uri_templates.get(
                None, {})

            for name, href in six.iteritems(unassigned_templates):
                templates[name] = href

            for entry in self.walk_resources(self, base_href):
                templates[entry.name] = entry.list_href

                if entry.is_list:
                    list_templates = self._registered_uri_templates.get(
                        entry.resource, {})

                    for name, href in six.iteritems(list_templates):
                        templates[name] = '%s%s' % (entry.list_href, href)

            self._uri_templates[base_href] = templates

        return self._uri_templates[base_href]

    @classmethod
    def walk_resources(cls, resource, list_href):
        """Yield all URI endpoints associated with a specified resource.

        Args:
            resource djblets.webapi.resources.WebAPIResource:
                The starting point for searching the resource tree.

            list_href (unicode):
                The path to the list resource, relative to the WebAPIResource
                provided. Used as a component of the URL in the API.

        Yields:
            RootResource.ResourceEntry:
            Resource entries for all sub-resources.
        """
        yield cls.ResourceEntry(name=resource.name_plural,
                                list_href=list_href,
                                resource=resource,
                                is_list=True)

        for child in resource.list_child_resources:
            child_href = '%s%s/' % (list_href, child.uri_name)

            for entry in cls.walk_resources(child, child_href):
                yield entry

        if resource.uri_object_key:
            object_href = '%s{%s}/' % (list_href, resource.uri_object_key)

            yield cls.ResourceEntry(name=resource.name,
                                    list_href=object_href,
                                    resource=resource,
                                    is_list=False)

            for child in resource.item_child_resources:
                child_href = '%s%s/' % (object_href, child.uri_name)

                for entry in cls.walk_resources(child, child_href):
                    yield entry

    def api_404_handler(self, request, api_format=None, *args, **kwargs):
        """Default handler at the end of the URL patterns.

        This returns an API 404, instead of a normal django 404.
        """
        return WebAPIResponseError(request,
                                   err=DOES_NOT_EXIST,
                                   api_format=api_format)

    def get_url_patterns(self):
        """Return the Django URL patterns for this object and its children.

        This returns the same list as
        :py:meth:`WebAPIResource.get_url_patterns`, but also introduces a
        generic catch-all 404 handler which returns API errors instead of HTML.
        """
        urlpatterns = super(RootResource, self).get_url_patterns()
        urlpatterns += never_cache_patterns(url(r'.*', self.api_404_handler))

        return urlpatterns

    def register_uri_template(self, name, relative_path,
                              relative_resource=None):
        """Register the specified resource for URI template serialization.

        This adds the specified name and relative resource to the Root
        Resource's URI templates.

        Args:
            name (unicode):
                The name of the associated resource being added to templates.

            relative_path (unicode):
                The path of the API resource relative to its parent resources.

            relative_resource (djblets.extensions.resources.ExtensionResource,
                               optional):
                The resource instance associated with this URI template.
        """
        # Clear the cache so that new lookups can detect newly added templates.
        self._uri_templates = {}

        templates = self._registered_uri_templates.setdefault(
            relative_resource, {})
        templates[name] = relative_path

    def unregister_uri_template(self, name, relative_resource=None):
        """Unregister the specified resource for URI template serialization.

        This removes the specified name and relative resource to the Root
        Resource's URI templates.

        Args:
            name (unicode):
                The name of the resource being removed from templates.

            relative_resource (djblets.extensions.resources.ExtensionResource,
                               optional):
                The resource instance associated with this URI template.
        """
        # Clear the cache so that new lookups can detect newly added templates.
        self._uri_templates = {}

        try:
            del self._registered_uri_templates[relative_resource][name]
        except KeyError:
            pass



"""Base support for consent types, data, and requirements."""

from __future__ import unicode_literals

try:
    # Python >= 3.4
    from enum import Enum
except ImportError:
    # Python < 3.4
    Enum = object

import dateutil.parser
from django.utils import timezone
from django.utils.translation import ugettext_lazy as _


class Consent(Enum):
    """Values for representing consent decisions."""

    #: Consent has not been set.
    UNSET = 0

    #: Consent was granted.
    GRANTED = 1

    #: Consent was denied.
    DENIED = 2


class ConsentData(object):
    """Data representing a granted or denied consent for a requirement.

    This tracks data about granted or denied consent for use in both setting
    a consent decision and for recording in an audit trail. The data tracks
    a number of important bits of information for later proving that consent
    was granted or denied at a particular point in time.

    Note that this does not track unset consent. It's only used when a decision
    was specifically made.

    Attributes:
        granted (bool):
            Whether consent was granted (``True``) or denied (``False``).

        extra_data (dict):
            Additional data to include that may be relevant for an audit.
            This can contain any information, and is up to the caller to
            determine.

        requirement_id (unicode):
            The ID of a :py:class:`ConsentRequirement` for which consent was
            granted or denied.

        source (unicode):
            The source where consent was decided. This is free-form text,
            and may represent a URL, API endpoint, or include other additional
            source-identifying information.

        timestamp (datetime.datetime):
            The date/time when the consent was decided.
    """

    @classmethod
    def parse_audit_info(cls, requirement_id, data):
        """Parse stored audit information.

        This data may come from the database or another tracking store.

        Args:
            requirement_id (unicode):
                The ID of a :py:class:`ConsentRequirement` for which consent
                was granted or denied.

            data (dict):
                The deserialized data to parse.

        Returns:
            ConsentData:
            The resulting consent data from the audit log.
        """
        return cls(requirement_id=requirement_id,
                   granted=data['granted'],
                   timestamp=dateutil.parser.parse(data['timestamp']),
                   source=data.get('source'),
                   extra_data=data.get('extra_data'))

    def __init__(self, requirement_id, granted=False, timestamp=None,
                 source=None, extra_data=None):
        """Initialize the consent data.

        Args:
            requirement_id (unicode):
                The ID of a :py:class:`ConsentRequirement` for which consent
                was granted or denied.

            granted (bool, optional):
                Whether consent was granted (``True``) or denied (``False``).
                Defaults to denied.

            timestamp (datetime.datetime, optional):
                The date/time when the consent was decided. If not specified,
                the current date/time will be used.

            source (unicode, optional):
                The source where consent was decided. This is free-form text,
                and may represent a URL, API endpoint, or include other
                additional source-identifying information.

            extra_data (dict, optional):
                Additional data to include that may be relevant for an audit.
                This can contain any information, and is up to the caller to
                determine.
        """
        self.requirement_id = requirement_id
        self.granted = granted
        self.timestamp = timestamp or timezone.now()
        self.source = source
        self.extra_data = extra_data

    def serialize_audit_info(self, identifier):
        """Serialize the consent data for audit tracking.

        Args:
            identifier (unicode):
                The identifier to store along with the rest of the data,
                to help identify the data subject at that point in time. This
                is considered opaque, and is up to the consent tracker.

        Returns:
            dict:
            Serialized data for the audit log.
        """
        result = {
            'identifier': identifier,
            'granted': self.granted,
            'timestamp': self.timestamp.isoformat(),
        }

        if self.source:
            result['source'] = self.source

        if self.extra_data:
            result['extra_data'] = self.extra_data

        return result


class BaseConsentRequirement(object):
    """Represents an aspect of the product requiring consent.

    For every piece of a product that requires consent (sending data to a
    tracking service, for instance), a Requirement must be registered in the
    :py:class:`registry
    <djblets.privacy.consent.registry.ConsentRequirementRegistry>`.

    This lists information on the requirement, describing it and optionally
    providing a URL for learning more about it (which may point to a privacy
    policy).

    It also contains helper functions to perform the lookups and generate
    consent data for this requirement.
    """

    #: The unique ID of the requirement.
    requirement_id = None

    #: The name of the requirement.
    name = None

    #: A brief summary of the requirement.
    summary = None

    #: A short description of why the requirement is needed.
    intent_description = None

    #: A short description of what data will be sent and how it will be used.
    data_use_description = None

    #: Text to use for the allow action.
    allow_text = _('Allow')

    #: Text to use for when the allow action is selected.
    allowed_text = _('Allowed')

    #: Text to use for the block action.
    block_text = _('Block')

    #: Text to use for when the block action is selected.
    blocked_text = _('Blocked')

    #: The icons used to represent this service or topic.
    #:
    #: This is in the form of a dictionary of icon resolution indicators (for
    #: ``srcset`` values) to URLs or relative static media paths.
    icons = {}

    def get_consent(self, user):
        """Return the user's consent decision for this requirement.

        Args:
            user (django.contrib.auth.models.User):
                The user who made a decision on consent for this requirement.

        Returns:
            djblets.privacy.consent.base.Consent:
            The user's consent decision for this requirement.
        """
        from djblets.privacy.consent.tracker import get_consent_tracker

        return get_consent_tracker().get_consent(user, self.requirement_id)

    def build_consent_data(self, **kwargs):
        """Returns ConsentData for this requirement.

        Args:
            **kwargs (dict):
                Positional arguments to pass to :py:class:`ConsentData`.

        Returns:
            ConsentData:
            The resulting consent data for this requirement and the specified
            arguments.
        """
        return ConsentData(requirement_id=self.requirement_id, **kwargs)
